@@ -40,15 +40,15 @@ struct connman_nat {
 	char *interface;
 };
 
-static int enable_ip_forward(connman_bool_t enable)
+static int enable_ip_forward(bool enable)
 {
 	FILE *f;
 
 	f = fopen("/proc/sys/net/ipv4/ip_forward", "r+");
-	if (f == NULL)
+	if (!f)
 		return -errno;
 
-	if (enable == TRUE)
+	if (enable)
 		fprintf(f, "1");
 	else
 		fprintf(f, "0");
@@ -66,7 +66,7 @@ static int enable_nat(struct connman_nat *nat)
 	g_free(nat->interface);
 	nat->interface = g_strdup(default_interface);
 
-	if (nat->interface == NULL)
+	if (!nat->interface)
 		return 0;
 
 	/* Enable masquerading */
@@ -86,7 +86,7 @@ static int enable_nat(struct connman_nat *nat)
 
 static void disable_nat(struct connman_nat *nat)
 {
-	if (nat->interface == NULL)
+	if (!nat->interface)
 		return;
 
 	/* Disable masquerading */
@@ -100,17 +100,17 @@ int __connman_nat_enable(const char *name, const char *address,
 	int err;
 
 	if (g_hash_table_size(nat_hash) == 0) {
-		err = enable_ip_forward(TRUE);
+		err = enable_ip_forward(true);
 		if (err < 0)
 			return err;
 	}
 
 	nat = g_try_new0(struct connman_nat, 1);
-	if (nat == NULL)
+	if (!nat)
 		goto err;
 
 	nat->fw = __connman_firewall_create();
-	if (nat->fw == NULL)
+	if (!nat->fw)
 		goto err;
 
 	nat->address = g_strdup(address);
@@ -121,11 +121,11 @@ int __connman_nat_enable(const char *name, const char *address,
 	return enable_nat(nat);
 
 err:
-	if (nat != NULL)
+	if (nat)
 		__connman_firewall_destroy(nat->fw);
 
 	if (g_hash_table_size(nat_hash) == 0)
-		enable_ip_forward(FALSE);
+		enable_ip_forward(false);
 
 	return -ENOMEM;
 }
@@ -135,7 +135,7 @@ void __connman_nat_disable(const char *name)
 	struct connman_nat *nat;
 
 	nat = g_hash_table_lookup(nat_hash, name);
-	if (nat == NULL)
+	if (!nat)
 		return;
 
 	disable_nat(nat);
@@ -143,7 +143,7 @@ void __connman_nat_disable(const char *name)
 	g_hash_table_remove(nat_hash, name);
 
 	if (g_hash_table_size(nat_hash) == 0)
-		enable_ip_forward(FALSE);
+		enable_ip_forward(false);
 }
 
 static void update_default_interface(struct connman_service *service)
@@ -162,7 +162,7 @@ static void update_default_interface(struct connman_service *service)
 
 	g_hash_table_iter_init(&iter, nat_hash);
 
-	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
+	while (g_hash_table_iter_next(&iter, &key, &value)) {
 		const char *name = key;
 		struct connman_nat *nat = value;
 
